@@ -1,6 +1,6 @@
 """Arbitrary precision arithmetic."""
 
-__version__ = '1.0.1'
+__version__ = '2.0.0'
 
 try:
     from ._rithm import (Fraction,
@@ -73,6 +73,10 @@ except ImportError:
         def __hash__(self) -> int:
             return hash(self._value)
 
+        def __invert__(self) -> 'Int':
+            return Int(~self._value,
+                       _parse=False)
+
         def __le__(self, other: 'Int') -> bool:
             return (self._value <= other._value
                     if isinstance(other, Int)
@@ -99,6 +103,14 @@ except ImportError:
             return Int(-self._value,
                        _parse=False)
 
+        def __pow__(self, exponent: 'Int') -> _Union['Fraction', 'Int']:
+            return ((Int(self._value ** exponent._value,
+                         _parse=False)
+                     if exponent >= _ZERO
+                     else Fraction(_ONE, self) ** -exponent)
+                    if isinstance(exponent, Int)
+                    else NotImplemented)
+
         def __repr__(self) -> str:
             return f'rithm.Int(\'{self._value}\')'
 
@@ -215,6 +227,19 @@ except ImportError:
             return Fraction(-self.numerator, self.denominator,
                             _normalize=False)
 
+        def __pow__(self, exponent: 'Int') -> 'Fraction':
+            return (
+                (Fraction(self.numerator ** exponent,
+                          self.denominator ** exponent,
+                          _normalize=False)
+                 if exponent >= _ZERO
+                 else Fraction(
+                    *_normalize_components_sign(self.denominator ** -exponent,
+                                                self.numerator ** -exponent),
+                    _normalize=False))
+                if isinstance(exponent, Int)
+                else NotImplemented)
+
         def __repr__(self) -> str:
             return f'rithm.Fraction({self.numerator!r}, {self.denominator!r})'
 
