@@ -1,12 +1,13 @@
 use std::cmp::Ordering;
 use std::convert::TryFrom;
-use std::{f64, fmt};
+use std::f64;
 
 use crate::traits::{
     AssigningAdditiveMonoid, AssigningBitwiseConjunctiveMagma, AssigningBitwiseDisjunctiveMonoid,
-    AssigningMultiplicativeMonoid, AssigningShiftingLeftMonoid, AssigningShiftingRightMonoid,
-    AssigningSubtractiveMagma, DivisivePartialMagma, DoublePrecision, DoublePrecisionOf,
-    ModularPartialMagma, ModularSubtractiveMagma, Oppose, OppositionOf, Unitary, Zeroable,
+    AssigningDivisivePartialMagma, AssigningMultiplicativeMonoid, AssigningShiftingLeftMonoid,
+    AssigningShiftingRightMonoid, AssigningSubtractiveMagma, DivisivePartialMagma, DoublePrecision,
+    DoublePrecisionOf, ModularPartialMagma, ModularSubtractiveMagma, Oppose, OppositionOf, Unitary,
+    Zeroable,
 };
 use crate::utils;
 
@@ -23,7 +24,7 @@ pub trait BinaryDigit = AssigningAdditiveMonoid
 
 pub(crate) type Sign = i8;
 
-pub fn binary_digits_to_base<SourceDigit, TargetDigit>(
+pub(crate) fn binary_digits_to_base<SourceDigit, TargetDigit>(
     source_digits: &[SourceDigit],
     source_shift: usize,
     target_base: usize,
@@ -37,28 +38,26 @@ where
         + TryFrom<DoublePrecisionOf<TargetDigit>>
         + Zeroable,
     DoublePrecisionOf<SourceDigit>: BinaryDigit,
-    DoublePrecisionOf<TargetDigit>: BinaryDigit
-        + DivisivePartialMagma
+    DoublePrecisionOf<TargetDigit>: AssigningDivisivePartialMagma
+        + BinaryDigit
         + From<SourceDigit>
         + From<TargetDigit>
         + ModularPartialMagma
         + TryFrom<usize>,
-    <DoublePrecisionOf<TargetDigit> as TryFrom<usize>>::Error: fmt::Debug,
-    <TargetDigit as TryFrom<DoublePrecisionOf<TargetDigit>>>::Error: fmt::Debug,
     usize: TryFrom<SourceDigit>,
 {
     if target_base & (target_base - 1) == 0 {
         binary_digits_to_binary_base(
             source_digits,
             source_shift,
-            utils::floor_log2::<usize>(target_base).unwrap(),
+            utils::floor_log2::<usize>(target_base),
         )
     } else {
         binary_digits_to_non_binary_base(source_digits, source_shift, target_base)
     }
 }
 
-pub fn digits_to_binary_base<SourceDigit, TargetDigit>(
+pub(crate) fn digits_to_binary_base<SourceDigit, TargetDigit>(
     source_digits: &[SourceDigit],
     source_base: usize,
     target_shift: usize,
@@ -80,7 +79,7 @@ where
     if source_base & (source_base - 1) == 0 {
         binary_digits_to_binary_base(
             source_digits,
-            utils::floor_log2::<usize>(source_base).unwrap(),
+            utils::floor_log2::<usize>(source_base),
             target_shift,
         )
     } else {
@@ -129,19 +128,18 @@ fn binary_digits_to_non_binary_base<SourceDigit, TargetDigit>(
 where
     SourceDigit: Copy + DoublePrecision,
     TargetDigit: Copy + DoublePrecision + TryFrom<DoublePrecisionOf<TargetDigit>> + Zeroable,
-    DoublePrecisionOf<TargetDigit>: BinaryDigit
-        + DivisivePartialMagma
+    DoublePrecisionOf<TargetDigit>: AssigningDivisivePartialMagma
+        + BinaryDigit
         + From<SourceDigit>
         + From<TargetDigit>
         + ModularPartialMagma
         + TryFrom<usize>,
-    <DoublePrecisionOf<TargetDigit> as TryFrom<usize>>::Error: fmt::Debug,
-    <TargetDigit as TryFrom<DoublePrecisionOf<TargetDigit>>>::Error: fmt::Debug,
 {
     let result_max_digits_count: usize = 1
         + ((((source_digits.len() * source_shift) as f64) / (target_base as f64).log2()) as usize);
     let mut result = Vec::<TargetDigit>::with_capacity(result_max_digits_count);
-    let target_base = DoublePrecisionOf::<TargetDigit>::try_from(target_base).unwrap();
+    let target_base =
+        unsafe { DoublePrecisionOf::<TargetDigit>::try_from(target_base).unwrap_unchecked() };
     for source_digit in source_digits.iter().rev() {
         let mut digit: DoublePrecisionOf<TargetDigit> =
             DoublePrecisionOf::<TargetDigit>::from(*source_digit);
@@ -149,11 +147,14 @@ where
             let step: DoublePrecisionOf<TargetDigit> =
                 (DoublePrecisionOf::<TargetDigit>::from(*result_position) << source_shift) | digit;
             digit = step / target_base;
-            *result_position = TargetDigit::try_from(step - digit * target_base).unwrap();
+            *result_position =
+                unsafe { TargetDigit::try_from(step - digit * target_base).unwrap_unchecked() };
         }
         while !digit.is_zero() {
-            result.push(TargetDigit::try_from(digit % target_base).unwrap());
-            digit = digit / target_base;
+            result.push(unsafe {
+                TargetDigit::try_from(digit.rem_euclid(target_base)).unwrap_unchecked()
+            });
+            digit /= target_base;
         }
     }
     if result.is_empty() {
@@ -172,6 +173,7 @@ where
     TargetDigit: DoublePrecision + TryFrom<DoublePrecisionOf<TargetDigit>>,
     DoublePrecisionOf<TargetDigit>: BinaryDigit + From<SourceDigit>,
 {
+    debug_assert!(target_shift > source_shift && source_shift > 0);
     let target_digit_mask = to_digit_mask::<DoublePrecisionOf<TargetDigit>>(target_shift);
     let result_capacity: usize =
         (source_digits.len() * target_shift + (target_shift - 1)) / target_shift;
@@ -199,7 +201,7 @@ where
     result
 }
 
-fn binary_digits_to_lesser_binary_base<SourceDigit, TargetDigit>(
+pub(crate) fn binary_digits_to_lesser_binary_base<SourceDigit, TargetDigit>(
     source_digits: &[SourceDigit],
     source_shift: usize,
     target_shift: usize,
@@ -210,9 +212,10 @@ where
     DoublePrecisionOf<SourceDigit>: BinaryDigit,
     usize: TryFrom<SourceDigit>,
 {
+    debug_assert!(source_shift > target_shift && target_shift > 0);
     let target_digit_mask = to_digit_mask::<DoublePrecisionOf<SourceDigit>>(target_shift);
     let result_digits_bits_count: usize = (source_digits.len() - 1) * source_shift
-        + utils::bit_length(*source_digits.last().unwrap());
+        + utils::bit_length(source_digits[source_digits.len() - 1]);
     let result_digits_count: usize = (result_digits_bits_count + (target_shift - 1)) / target_shift;
     let mut result = Vec::<TargetDigit>::with_capacity(result_digits_count);
     let mut accumulator = DoublePrecisionOf::<SourceDigit>::zero();
@@ -241,13 +244,207 @@ where
     result
 }
 
+pub(crate) fn checked_div<Digit, const SHIFT: usize>(
+    dividend: &[Digit],
+    dividend_sign: Sign,
+    divisor: &[Digit],
+    divisor_sign: Sign,
+) -> Option<(Sign, Vec<Digit>)>
+where
+    Digit: BinaryDigit
+        + DoublePrecision
+        + From<u8>
+        + ModularSubtractiveMagma
+        + Oppose
+        + TryFrom<DoublePrecisionOf<Digit>>
+        + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>>
+        + TryFrom<usize>,
+    DoublePrecisionOf<Digit>: BinaryDigit + DivisivePartialMagma + Oppose,
+    OppositionOf<Digit>:
+        BinaryDigit + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>> + TryFrom<Digit>,
+    OppositionOf<DoublePrecisionOf<Digit>>: BinaryDigit + From<Digit> + From<OppositionOf<Digit>>,
+    usize: TryFrom<Digit>,
+{
+    if divisor_sign.is_zero() {
+        None
+    } else if dividend_sign.is_zero() || digits_lesser_than(dividend, divisor) {
+        Some((Sign::zero(), vec![Digit::zero()]))
+    } else if divisor.len() == 1 {
+        let (digits, _) = div_rem_digits_by_digit::<Digit, SHIFT>(dividend, divisor[0]);
+        Some((dividend_sign * divisor_sign, digits))
+    } else {
+        let (digits, _) = div_rem_two_or_more_digits::<Digit, SHIFT>(dividend, divisor);
+        Some((
+            dividend_sign * divisor_sign * ((digits.len() > 1 || !digits[0].is_zero()) as Sign),
+            digits,
+        ))
+    }
+}
+
+pub(crate) fn checked_div_euclid<Digit, const SHIFT: usize>(
+    dividend: &[Digit],
+    dividend_sign: Sign,
+    divisor: &[Digit],
+    divisor_sign: Sign,
+) -> Option<(Sign, Vec<Digit>)>
+where
+    Digit: BinaryDigit
+        + DoublePrecision
+        + From<u8>
+        + ModularSubtractiveMagma
+        + Oppose
+        + TryFrom<DoublePrecisionOf<Digit>>
+        + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>>
+        + TryFrom<usize>,
+    DoublePrecisionOf<Digit>: BinaryDigit + DivisivePartialMagma + Oppose,
+    OppositionOf<Digit>:
+        BinaryDigit + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>> + TryFrom<Digit>,
+    OppositionOf<DoublePrecisionOf<Digit>>: BinaryDigit + From<Digit> + From<OppositionOf<Digit>>,
+    usize: TryFrom<Digit>,
+{
+    if divisor_sign.is_zero() {
+        None
+    } else if dividend_sign.is_zero() {
+        Some((Sign::zero(), vec![Digit::zero()]))
+    } else if digits_lesser_than(dividend, divisor) {
+        Some(
+            if (dividend_sign.is_negative() && divisor_sign.is_positive())
+                || (dividend_sign.is_positive() && divisor_sign.is_negative())
+            {
+                (-Sign::one(), vec![Digit::one()])
+            } else {
+                (Sign::zero(), vec![Digit::zero()])
+            },
+        )
+    } else {
+        let (sign, mut digits, remainder_is_non_zero) = if divisor.len() == 1 {
+            let (digits, remainder_digit) =
+                div_rem_digits_by_digit::<Digit, SHIFT>(dividend, divisor[0]);
+            (
+                dividend_sign * divisor_sign,
+                digits,
+                !remainder_digit.is_zero(),
+            )
+        } else {
+            let (digits, remainder_digits) =
+                div_rem_two_or_more_digits::<Digit, SHIFT>(dividend, divisor);
+            (
+                dividend_sign * divisor_sign * ((digits.len() > 1 || !digits[0].is_zero()) as Sign),
+                digits,
+                remainder_digits.len() > 1 || !remainder_digits[0].is_zero(),
+            )
+        };
+        if remainder_is_non_zero
+            && ((dividend_sign.is_negative() && divisor_sign.is_positive())
+                || (dividend_sign.is_positive() && divisor_sign.is_negative()))
+        {
+            digits = sum_digits::<Digit, SHIFT>(&digits, &[Digit::one()]);
+        }
+        Some((sign, digits))
+    }
+}
+
+pub(crate) fn checked_rem<Digit, const SHIFT: usize>(
+    dividend: &[Digit],
+    dividend_sign: Sign,
+    divisor: &[Digit],
+    divisor_sign: Sign,
+) -> Option<(Sign, Vec<Digit>)>
+where
+    Digit: BinaryDigit
+        + DoublePrecision
+        + From<u8>
+        + ModularSubtractiveMagma
+        + Oppose
+        + TryFrom<DoublePrecisionOf<Digit>>
+        + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>>
+        + TryFrom<usize>,
+    DoublePrecisionOf<Digit>: BinaryDigit + DivisivePartialMagma + Oppose,
+    OppositionOf<Digit>:
+        BinaryDigit + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>> + TryFrom<Digit>,
+    OppositionOf<DoublePrecisionOf<Digit>>: BinaryDigit + From<Digit> + From<OppositionOf<Digit>>,
+    usize: TryFrom<Digit>,
+{
+    if divisor_sign.is_zero() {
+        None
+    } else if dividend_sign.is_zero() || digits_lesser_than(dividend, divisor) {
+        Some((dividend_sign, dividend.to_vec()))
+    } else if divisor.len() == 1 {
+        let (_, digit) = div_rem_digits_by_digit::<Digit, SHIFT>(dividend, divisor[0]);
+        Some((dividend_sign * ((!digit.is_zero()) as Sign), vec![digit]))
+    } else {
+        let (_, digits) = div_rem_two_or_more_digits::<Digit, SHIFT>(dividend, divisor);
+        Some((
+            dividend_sign * ((digits.len() > 1 || !digits[0].is_zero()) as Sign),
+            digits,
+        ))
+    }
+}
+
+pub(crate) fn checked_rem_euclid<Digit, const SHIFT: usize>(
+    dividend: &[Digit],
+    dividend_sign: Sign,
+    divisor: &[Digit],
+    divisor_sign: Sign,
+) -> Option<(Sign, Vec<Digit>)>
+where
+    Digit: BinaryDigit
+        + DoublePrecision
+        + From<u8>
+        + ModularSubtractiveMagma
+        + Oppose
+        + TryFrom<DoublePrecisionOf<Digit>>
+        + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>>
+        + TryFrom<usize>,
+    DoublePrecisionOf<Digit>: BinaryDigit + DivisivePartialMagma + Oppose,
+    OppositionOf<Digit>:
+        BinaryDigit + TryFrom<OppositionOf<DoublePrecisionOf<Digit>>> + TryFrom<Digit>,
+    OppositionOf<DoublePrecisionOf<Digit>>: BinaryDigit + From<Digit> + From<OppositionOf<Digit>>,
+    usize: TryFrom<Digit>,
+{
+    if divisor_sign.is_zero() {
+        None
+    } else if dividend_sign.is_zero() {
+        Some((dividend_sign, dividend.to_vec()))
+    } else if digits_lesser_than(dividend, divisor) {
+        Some(
+            if (dividend_sign.is_negative() && divisor_sign.is_positive())
+                || (dividend_sign.is_positive() && divisor_sign.is_negative())
+            {
+                let mut sign = dividend_sign;
+                let digits = subtract_digits::<Digit, SHIFT>(dividend, divisor, &mut sign);
+                (sign, digits)
+            } else {
+                (dividend_sign, dividend.to_vec())
+            },
+        )
+    } else {
+        let (mut sign, mut digits) = if divisor.len() == 1 {
+            let (_, digit) = div_rem_digits_by_digit::<Digit, SHIFT>(dividend, divisor[0]);
+            (dividend_sign * ((!digit.is_zero()) as Sign), vec![digit])
+        } else {
+            let (_, digits) = div_rem_two_or_more_digits::<Digit, SHIFT>(dividend, divisor);
+            (
+                dividend_sign * ((digits.len() > 1 || !digits[0].is_zero()) as Sign),
+                digits,
+            )
+        };
+        if (divisor_sign.is_negative() && sign.is_positive())
+            || (divisor_sign.is_positive() && sign.is_negative())
+        {
+            digits = subtract_digits::<Digit, SHIFT>(&digits, divisor, &mut sign);
+        }
+        Some((sign, digits))
+    }
+}
+
 #[inline]
-pub fn digits_lesser_than<Digit: PartialOrd>(left: &[Digit], right: &[Digit]) -> bool {
+pub(crate) fn digits_lesser_than<Digit: PartialOrd>(left: &[Digit], right: &[Digit]) -> bool {
     left.len() < right.len()
         || left.len() == right.len() && left.iter().rev().lt(right.iter().rev())
 }
 
-pub fn divrem_digits_by_digit<Digit, const SHIFT: usize>(
+pub(crate) fn div_rem_digits_by_digit<Digit, const SHIFT: usize>(
     dividend: &[Digit],
     divisor: Digit,
 ) -> (Vec<Digit>, Digit)
@@ -272,7 +469,7 @@ where
     })
 }
 
-pub fn divrem_two_or_more_digits<Digit, const SHIFT: usize>(
+pub(crate) fn div_rem_two_or_more_digits<Digit, const SHIFT: usize>(
     dividend: &[Digit],
     divisor: &[Digit],
 ) -> (Vec<Digit>, Vec<Digit>)
@@ -464,7 +661,7 @@ where
     result
 }
 
-pub fn multiply_digits<Digit, const SEPARATOR: char, const SHIFT: usize>(
+pub(crate) fn multiply_digits<Digit, const SHIFT: usize>(
     first: &[Digit],
     second: &[Digit],
 ) -> Vec<Digit>
@@ -496,11 +693,11 @@ where
         return if size_shortest == 1 && shortest[0].is_zero() {
             vec![Digit::zero()]
         } else {
-            multiply_digits_plain::<Digit, SEPARATOR, SHIFT>(*shortest, *longest)
+            multiply_digits_plain::<Digit, SHIFT>(*shortest, *longest)
         };
     };
     if 2 * size_shortest <= size_longest {
-        return multiply_digits_lopsided::<Digit, SEPARATOR, SHIFT>(*shortest, *longest);
+        return multiply_digits_lopsided::<Digit, SHIFT>(*shortest, *longest);
     }
     let shift = size_longest >> 1;
     let (shortest_high, shortest_low) = split_digits(*shortest, shift);
@@ -510,33 +707,30 @@ where
         split_digits(*longest, shift)
     };
     let mut result = vec![Digit::zero(); size_shortest + size_longest];
-    let highs_product = multiply_digits::<Digit, SEPARATOR, SHIFT>(&shortest_high, &longest_high);
+    let highs_product = multiply_digits::<Digit, SHIFT>(&shortest_high, &longest_high);
     for (index, &digit) in highs_product.iter().enumerate() {
         result[index + 2 * shift] = digit;
     }
-    let lows_product = multiply_digits::<Digit, SEPARATOR, SHIFT>(&shortest_low, &longest_low);
+    let lows_product = multiply_digits::<Digit, SHIFT>(&shortest_low, &longest_low);
     for (index, &digit) in lows_product.iter().enumerate() {
         result[index] = digit;
     }
-    subtract_digits_in_place::<Digit, SEPARATOR, SHIFT>(&mut result[shift..], &lows_product);
-    subtract_digits_in_place::<Digit, SEPARATOR, SHIFT>(&mut result[shift..], &highs_product);
-    let shortest_components_sum =
-        sum_digits::<Digit, SEPARATOR, SHIFT>(&shortest_high, &shortest_low);
+    subtract_digits_in_place::<Digit, SHIFT>(&mut result[shift..], &lows_product);
+    subtract_digits_in_place::<Digit, SHIFT>(&mut result[shift..], &highs_product);
+    let shortest_components_sum = sum_digits::<Digit, SHIFT>(&shortest_high, &shortest_low);
     let longest_components_sum = if shortest.as_ptr() == longest.as_ptr() {
         shortest_components_sum.clone()
     } else {
-        sum_digits::<Digit, SEPARATOR, SHIFT>(&longest_high, &longest_low)
+        sum_digits::<Digit, SHIFT>(&longest_high, &longest_low)
     };
-    let components_sums_product = multiply_digits::<Digit, SEPARATOR, SHIFT>(
-        &shortest_components_sum,
-        &longest_components_sum,
-    );
-    sum_digits_in_place::<Digit, SEPARATOR, SHIFT>(&mut result[shift..], &components_sums_product);
+    let components_sums_product =
+        multiply_digits::<Digit, SHIFT>(&shortest_components_sum, &longest_components_sum);
+    sum_digits_in_place::<Digit, SHIFT>(&mut result[shift..], &components_sums_product);
     normalize_digits(&mut result);
     result
 }
 
-fn multiply_digits_lopsided<Digit, const SEPARATOR: char, const SHIFT: usize>(
+fn multiply_digits_lopsided<Digit, const SHIFT: usize>(
     shortest: &[Digit],
     longest: &[Digit],
 ) -> Vec<Digit>
@@ -554,14 +748,11 @@ where
     let mut processed_digits_count = 0;
     while size_longest > 0 {
         let step_digits_count = size_longest.min(size_shortest);
-        let product = multiply_digits::<Digit, SEPARATOR, SHIFT>(
+        let product = multiply_digits::<Digit, SHIFT>(
             shortest,
             &longest[processed_digits_count..processed_digits_count + step_digits_count].to_vec(),
         );
-        sum_digits_in_place::<Digit, SEPARATOR, SHIFT>(
-            &mut result[processed_digits_count..],
-            &product,
-        );
+        sum_digits_in_place::<Digit, SHIFT>(&mut result[processed_digits_count..], &product);
         size_longest -= step_digits_count;
         processed_digits_count += step_digits_count;
     }
@@ -569,7 +760,7 @@ where
     result
 }
 
-fn multiply_digits_plain<Digit, const SEPARATOR: char, const SHIFT: usize>(
+fn multiply_digits_plain<Digit, const SHIFT: usize>(
     shortest: &[Digit],
     longest: &[Digit],
 ) -> Vec<Digit>
@@ -579,7 +770,7 @@ where
 {
     let size_shortest = shortest.len();
     let size_longest = longest.len();
-    let mut result: Vec<Digit> = vec![Digit::zero(); size_shortest + size_longest];
+    let mut result = vec![Digit::zero(); size_shortest + size_longest];
     let digit_mask = to_digit_mask::<DoublePrecisionOf<Digit>>(SHIFT);
     if shortest.as_ptr() == longest.as_ptr() {
         for index in 0..size_shortest {
@@ -687,13 +878,42 @@ where
     (high, low)
 }
 
-pub fn subtract_digits<Digit, const SEPARATOR: char, const SHIFT: usize>(
+pub(crate) fn subtract_signed_digits<Digit, const SHIFT: usize>(
+    minuend: &[Digit],
+    minuend_sign: Sign,
+    subtrahend: &[Digit],
+    subtrahend_sign: Sign,
+) -> (Sign, Vec<Digit>)
+where
+    Digit: BinaryDigit + ModularSubtractiveMagma,
+{
+    if minuend_sign.is_negative() {
+        if subtrahend_sign.is_negative() {
+            let mut sign = Sign::one();
+            let digits = subtract_digits::<Digit, SHIFT>(subtrahend, minuend, &mut sign);
+            (sign, digits)
+        } else {
+            (
+                -Sign::one(),
+                sum_digits::<Digit, SHIFT>(minuend, subtrahend),
+            )
+        }
+    } else if subtrahend_sign.is_negative() {
+        (Sign::one(), sum_digits::<Digit, SHIFT>(minuend, subtrahend))
+    } else {
+        let mut sign = Sign::one();
+        let digits = subtract_digits::<Digit, SHIFT>(minuend, subtrahend, &mut sign);
+        (sign, digits)
+    }
+}
+
+fn subtract_digits<Digit, const SHIFT: usize>(
     first: &[Digit],
     second: &[Digit],
     sign: &mut Sign,
 ) -> Vec<Digit>
 where
-    Digit: BinaryDigit + ModularSubtractiveMagma + TryFrom<usize>,
+    Digit: BinaryDigit + ModularSubtractiveMagma,
 {
     let mut longest = &first;
     let mut shortest = &second;
@@ -745,7 +965,7 @@ where
     result
 }
 
-fn subtract_digits_in_place<Digit, const SEPARATOR: char, const SHIFT: usize>(
+fn subtract_digits_in_place<Digit, const SHIFT: usize>(
     longest: &mut [Digit],
     shortest: &[Digit],
 ) -> Digit
@@ -770,12 +990,38 @@ where
     accumulator
 }
 
-pub fn sum_digits<Digit, const SEPARATOR: char, const SHIFT: usize>(
+pub(crate) fn sum_signed_digits<Digit, const SHIFT: usize>(
     first: &[Digit],
+    first_sign: Sign,
     second: &[Digit],
-) -> Vec<Digit>
+    second_sign: Sign,
+) -> (Sign, Vec<Digit>)
 where
-    Digit: BinaryDigit + TryFrom<usize>,
+    Digit: BinaryDigit + ModularSubtractiveMagma,
+{
+    if first_sign.is_negative() {
+        if second_sign.is_negative() {
+            (-Sign::one(), sum_digits::<Digit, SHIFT>(first, second))
+        } else {
+            let mut sign = Sign::one();
+            let digits = subtract_digits::<Digit, SHIFT>(second, first, &mut sign);
+            (sign, digits)
+        }
+    } else if second_sign.is_negative() {
+        let mut sign = Sign::one();
+        let digits = subtract_digits::<Digit, SHIFT>(first, second, &mut sign);
+        (sign, digits)
+    } else {
+        (
+            first_sign.max(second_sign),
+            sum_digits::<Digit, SHIFT>(first, second),
+        )
+    }
+}
+
+fn sum_digits<Digit, const SHIFT: usize>(first: &[Digit], second: &[Digit]) -> Vec<Digit>
+where
+    Digit: BinaryDigit,
 {
     let mut longest = &first;
     let mut shortest = &second;
@@ -803,7 +1049,7 @@ where
     result
 }
 
-fn sum_digits_in_place<Digit, const SEPARATOR: char, const SHIFT: usize>(
+fn sum_digits_in_place<Digit, const SHIFT: usize>(
     longest: &mut [Digit],
     shortest: &[Digit],
 ) -> Digit
@@ -828,7 +1074,7 @@ where
     accumulator
 }
 
-pub fn normalize_digits<Digit>(digits: &mut Vec<Digit>)
+pub(crate) fn normalize_digits<Digit>(digits: &mut Vec<Digit>)
 where
     Digit: Clone + Zeroable,
 {
@@ -841,7 +1087,7 @@ where
     }
 }
 
-pub fn reduce_digits<Digit, Output, const SHIFT: usize>(digits: &[Digit]) -> Output
+pub(crate) fn reduce_digits<Digit, Output, const SHIFT: usize>(digits: &[Digit]) -> Output
 where
     Digit: Copy,
     Output: BinaryDigit + From<Digit>,
@@ -854,7 +1100,7 @@ where
 }
 
 #[inline]
-pub fn to_digit_mask<Digit>(shift: usize) -> Digit
+pub(crate) fn to_digit_mask<Digit>(shift: usize) -> Digit
 where
     Digit: BinaryDigit,
 {
