@@ -1,14 +1,19 @@
 #![feature(destructuring_assignment)]
 #![feature(option_result_unwrap_unchecked)]
+#![feature(specialization)]
 #![feature(trait_alias)]
 
-use crate::traits::{CheckedDiv, Gcd, Modular, Oppositive, Unitary, Zeroable};
+use crate::traits::{
+    Abs, CheckedDiv, CheckedDivEuclid, CheckedPow, CheckedRemEuclid, FromStrRadix, Gcd, Oppositive,
+    Pow, Unitary, Zeroable,
+};
 use pyo3::basic::CompareOp;
 use pyo3::class::PyObjectProtocol;
 use pyo3::exceptions::*;
 use pyo3::ffi::Py_hash_t;
 use pyo3::prelude::{pyclass, pymethods, pymodule, pyproto, PyModule, PyResult, Python};
-use pyo3::PyNumberProtocol;
+use pyo3::{IntoPy, PyNumberProtocol, PyObject};
+use std::convert::TryFrom;
 
 pub mod big_int;
 mod digits;
@@ -42,7 +47,7 @@ impl PyInt {
     fn new(_string: &str, base: u32) -> PyResult<Self> {
         match _BigInt::from_str_radix(_string, base) {
             Ok(value) => Ok(PyInt(value)),
-            Err(reason) => Err(PyValueError::new_err(reason)),
+            Err(reason) => Err(PyValueError::new_err(reason.to_string())),
         }
     }
 
@@ -92,7 +97,7 @@ impl PyNumberProtocol for PyInt {
     }
 
     fn __divmod__(lhs: PyInt, rhs: PyInt) -> PyResult<(PyInt, PyInt)> {
-        match divmod(lhs.0, rhs.0) {
+        match checked_div_rem_euclid(lhs.0, rhs.0) {
             Some((quotient, remainder)) => Ok((PyInt(quotient), PyInt(remainder))),
             None => Err(PyZeroDivisionError::new_err(
                 UNDEFINED_DIVISION_ERROR_MESSAGE,
@@ -101,16 +106,21 @@ impl PyNumberProtocol for PyInt {
     }
 
     fn __floordiv__(lhs: PyInt, rhs: PyInt) -> PyResult<PyInt> {
-        match divmod(lhs.0, rhs.0) {
-            Some((result, _)) => Ok(PyInt(result)),
+        match lhs.0.checked_div_euclid(rhs.0) {
+            Some(result) => Ok(PyInt(result)),
             None => Err(PyZeroDivisionError::new_err(
                 UNDEFINED_DIVISION_ERROR_MESSAGE,
             )),
         }
     }
+
+    fn __invert__(&self) -> PyInt {
+        PyInt(!self.0.clone())
+    }
+
     fn __mod__(lhs: PyInt, rhs: PyInt) -> PyResult<PyInt> {
-        match divmod(lhs.0, rhs.0) {
-            Some((_, result)) => Ok(PyInt(result)),
+        match lhs.0.checked_rem_euclid(rhs.0) {
+            Some(result) => Ok(PyInt(result)),
             None => Err(PyZeroDivisionError::new_err(
                 UNDEFINED_DIVISION_ERROR_MESSAGE,
             )),
@@ -125,6 +135,26 @@ impl PyNumberProtocol for PyInt {
         PyInt(-self.0.clone())
     }
 
+    fn __pow__(lhs: PyInt, rhs: PyInt, _modulo: Option<PyInt>) -> PyResult<PyObject> {
+        debug_assert!(_modulo.is_none());
+        Ok({
+            if rhs.0.is_negative() {
+                to_py_object(match unsafe {
+                    _Fraction::new(lhs.0, _BigInt::one()).unwrap_unchecked()
+                }
+                .checked_pow(rhs.0)
+                {
+                    Some(value) => Ok(PyFraction(value)),
+                    None => Err(PyZeroDivisionError::new_err(
+                        UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    )),
+                }?)
+            } else {
+                to_py_object(PyInt(lhs.0.pow(rhs.0)))
+            }
+        })
+    }
+
     fn __sub__(lhs: PyInt, rhs: PyInt) -> PyInt {
         PyInt(lhs.0 - rhs.0)
     }
@@ -139,17 +169,51 @@ impl PyNumberProtocol for PyInt {
     }
 }
 
-fn divmod(dividend: _BigInt, divisor: _BigInt) -> Option<(_BigInt, _BigInt)> {
-    let (mut quotient, mut modulo) = dividend.divrem(&divisor)?;
+fn checked_div_rem_euclid(dividend: _BigInt, divisor: _BigInt) -> Option<(_BigInt, _BigInt)> {
+    let (mut quotient, mut modulo) = dividend.checked_div_rem(&divisor)?;
     if (divisor.is_negative() && modulo.is_positive())
         || (divisor.is_positive() && modulo.is_negative())
     {
-        quotient = quotient - _BigInt::one();
-        modulo = modulo + divisor;
+        quotient -= _BigInt::one();
+        modulo += divisor;
     }
     Some((quotient, modulo))
 }
 
+fn hash(value: &_BigInt) -> usize {
+    #[cfg(target_arch = "x86")]
+    const HASH_BITS: usize = 31;
+    #[cfg(not(target_arch = "x86"))]
+    const HASH_BITS: usize = 61;
+    const HASH_MODULUS: usize = (1 << HASH_BITS) - 1;
+    if value.digits().len() == 1 {
+        return if value.is_negative() {
+            usize::MAX
+                - unsafe {
+                    usize::try_from(
+                        value.digits()[0] + <Digit as From<bool>>::from(value.digits()[0].is_one()),
+                    )
+                    .unwrap_unchecked()
+                }
+                + 1
+        } else {
+            unsafe { usize::try_from(value.digits()[0]).unwrap_unchecked() }
+        };
+    };
+    let mut result = 0;
+    for &position in value.digits().iter().rev() {
+        result = ((result << BINARY_SHIFT) & HASH_MODULUS) | (result >> (HASH_BITS - BINARY_SHIFT));
+        result += unsafe { usize::try_from(position).unwrap_unchecked() };
+        if result >= HASH_MODULUS {
+            result -= HASH_MODULUS;
+        }
+    }
+    if value.is_negative() {
+        result = usize::MAX - result + 1
+    };
+    result - ((result == usize::MAX) as usize)
+}
+
 #[pyproto]
 impl PyObjectProtocol for PyInt {
     fn __bool__(self) -> bool {
@@ -157,7 +221,7 @@ impl PyObjectProtocol for PyInt {
     }
 
     fn __hash__(&self) -> Py_hash_t {
-        self.0.hash() as Py_hash_t
+        hash(&self.0) as Py_hash_t
     }
 
     fn __repr__(&self) -> String {
@@ -173,29 +237,6 @@ impl PyObjectProtocol for PyInt {
     }
 }
 
-#[pyproto]
-impl PyObjectProtocol for PyFraction {
-    fn __bool__(self) -> bool {
-        self.numerator().__bool__()
-    }
-
-    fn __repr__(&self) -> String {
-        format!(
-            "rithm.Fraction({}, {})",
-            self.numerator().__repr__(),
-            self.denominator().__repr__()
-        )
-    }
-
-    fn __richcmp__(&self, other: PyFraction, op: CompareOp) -> bool {
-        compare(&self.0, &other.0, op)
-    }
-
-    fn __str__(&self) -> String {
-        self.0.to_string()
-    }
-}
-
 #[pyproto]
 impl PyNumberProtocol for PyFraction {
     fn __abs__(&self) -> PyFraction {
@@ -214,6 +255,16 @@ impl PyNumberProtocol for PyFraction {
         PyFraction(-self.0.clone())
     }
 
+    fn __pow__(lhs: PyFraction, rhs: PyInt, _modulo: Option<PyInt>) -> PyResult<PyFraction> {
+        debug_assert!(_modulo.is_none());
+        match lhs.0.checked_pow(rhs.0) {
+            Some(value) => Ok(PyFraction(value)),
+            None => Err(PyZeroDivisionError::new_err(
+                UNDEFINED_DIVISION_ERROR_MESSAGE,
+            )),
+        }
+    }
+
     fn __sub__(lhs: PyFraction, rhs: PyFraction) -> PyFraction {
         PyFraction(lhs.0 - rhs.0)
     }
@@ -228,6 +279,29 @@ impl PyNumberProtocol for PyFraction {
     }
 }
 
+#[pyproto]
+impl PyObjectProtocol for PyFraction {
+    fn __bool__(self) -> bool {
+        self.numerator().__bool__()
+    }
+
+    fn __repr__(&self) -> String {
+        format!(
+            "rithm.Fraction({}, {})",
+            self.numerator().__repr__(),
+            self.denominator().__repr__()
+        )
+    }
+
+    fn __richcmp__(&self, other: PyFraction, op: CompareOp) -> bool {
+        compare(&self.0, &other.0, op)
+    }
+
+    fn __str__(&self) -> String {
+        self.0.to_string()
+    }
+}
+
 fn compare<T: PartialOrd>(left: &T, right: &T, op: CompareOp) -> bool {
     match op {
         CompareOp::Eq => left == right,
@@ -247,3 +321,7 @@ fn _rithm(_py: Python, module: &PyModule) -> PyResult<()> {
     module.add_class::<PyFraction>()?;
     Ok(())
 }
+
+fn to_py_object<T: IntoPy<PyObject>>(result: T) -> PyObject {
+    Python::with_gil(|py| result.into_py(py))
+}
