@@ -19,3 +19,5 @@ non_zero_ints_with_non_zero_builtin_ints = ints_with_builtin_ints.filter(all)
 fractions_with_builtin_fractions = strategies.builds(
     to_fraction_with_builtin_fraction, ints_with_builtin_ints,
     non_zero_ints_with_non_zero_builtin_ints)
+small_ints_with_builtin_ints = strategies.builds(
+    to_int_with_builtin_int, strategies.integers(-100, 100).map(str))
